import functools
import logging
import os
import sys

import click
import coloredlogs
import wotemu.cli.app
import wotemu.cli.broker
import wotemu.cli.chaos
import wotemu.cli.compose
import wotemu.cli.limits
import wotemu.cli.report
import wotemu.cli.routes
import wotemu.cli.stop
import wotemu.cli.waiter
import wotemu.config

_COMMAND_KWARGS = {
    "context_settings": {
        "show_default": True
    }
}

_DEFAULT_DOCKER_SOCK = "unix://{}".format(wotemu.config.DEFAULT_DOCKER_SOCKET)

_logger = logging.getLogger(__name__)


def _catch(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            _logger.error("CLI error", exc_info=True)
            sys.exit(1)

    return wrapper


@click.group()
@click.option("--log-level", default="DEBUG")
@click.option("--quiet", is_flag=True)
@click.option("--root-logger", is_flag=True)
@click.pass_context
def cli(ctx, log_level, quiet, root_logger):
    """Root CLI command."""

    logging.getLogger().addHandler(logging.NullHandler())

    if not quiet:
        pkg_name = __name__.split(".")[0]
        cli_logger_name = None if root_logger else pkg_name
        cli_logger = logging.getLogger(cli_logger_name)
        coloredlogs.install(level=log_level, logger=cli_logger)

    wotemu.config.log_config()
    ctx.obj = wotemu.config.get_env_config()


@cli.command(**_COMMAND_KWARGS)
@click.option("--rtable-name", default="wotemu")
@click.option("--rtable-mark", type=int, default=1)
@click.option("--apply", is_flag=True)
@click.pass_obj
@_catch
def route(conf, **kwargs):
    """This command should be called in the initialization phase of all WoTemu nodes. 
    Updates the routing configuration of this container to force WoT communications to go through 
    the network gateway container. Requires access to the Docker daemon of a manager node."""

    wotemu.cli.routes.update_routing(conf, **kwargs)


@cli.command(**_COMMAND_KWARGS)
@click.option("--docker-url", default=_DEFAULT_DOCKER_SOCK)
@click.option("--netem", type=str, multiple=True)
@click.option("--duration", type=str, default="72h")
@click.pass_obj
@_catch
def chaos(conf, **kwargs):
    """This is the main command of network gateway containers. Runs and controls 
    a set of Pumba subprocesses that degrade the network stack of this container 
    to simulate real-life network conditions."""

    wotemu.cli.chaos.create_chaos(conf, **kwargs)


@cli.command(**_COMMAND_KWARGS)
@click.option("--path", required=True, type=str)
@click.option("--func", type=str, default="app")
@click.option('--func-param', multiple=True, type=(str, str))
@click.option("--hostname", type=str, default=None)
@click.option("--enable-http", is_flag=True)
@click.option("--enable-coap", is_flag=True)
@click.option("--enable-mqtt", is_flag=True)
@click.option("--enable-ws", is_flag=True)
@click.option("--disable-monitor", is_flag=True)
@click.pass_obj
@_catch
def app(conf, **kwargs):
    """Runs a user-defined WoT application injected with a decorated WoTPy entrypoint."""

    wotemu.cli.app.run_app(conf, **kwargs)


@cli.command(**_COMMAND_KWARGS)
@click.option("--path", required=True, type=str)
@click.option("--output", type=str, default=None)
@click.option("--func", type=str, default="topology")
@click.pass_obj
@_catch
def compose(conf, **kwargs):
    """Takes a topology definition and builds a Compose file for emulation."""

    wotemu.cli.compose.build_compose(conf, **kwargs)


@cli.command(**_COMMAND_KWARGS)
@click.option("--docker-url", default=_DEFAULT_DOCKER_SOCK)
@click.option("--speed", type=int, required=True)
@click.option("--timeout", type=int, default=300)
@click.option("--wait", type=int, default=2)
@click.option("--local", is_flag=True)
@click.pass_obj
@_catch
def limits(conf, **kwargs):
    """Updates the CPU performance limits from inside a WoTemu container."""

    wotemu.cli.limits.update_limits(conf, **kwargs)


@cli.command(**_COMMAND_KWARGS)
@click.option("--sleep", type=float, default=1.0)
@click.option("--timeout", type=float, default=60.0)
@click.option("--waiter-base", is_flag=True)
@click.option("--waiter-mqtt", is_flag=True)
@click.pass_obj
@_catch
def wait(conf, **kwargs):
    """Waits for the services of the WoTemu stack to be available."""

    wotemu.cli.waiter.wait_services(conf, **kwargs)


@cli.command(**_COMMAND_KWARGS)
@click.option("--disable-monitor", is_flag=True)
@click.pass_obj
@_catch
def broker(conf, **kwargs):
    """Runs a MQTT broker."""

    wotemu.cli.broker.run_mqtt_broker(conf, **kwargs)


@cli.command(**_COMMAND_KWARGS)
@click.option("--compose-file", required=True)
@click.option("--stack", required=True)
@click.option("--redis-url", default=None)
@click.option("--tail", type=int, default=60)
@click.pass_obj
@_catch
def stop(conf, **kwargs):
    """Stops an emulation stack that is currently active. 
    Unlike 'docker stack rm' this keeps the in-memory traces 
    if necessary and logs other events of interest."""

    wotemu.cli.stop.stop_stack(conf, **kwargs)


@cli.command(**_COMMAND_KWARGS)
@click.option("--out", required=True)
@click.option("--stack", default=None)
@click.option("--redis-url", default=None)
@click.option("--json", is_flag=True)
@click.pass_obj
@_catch
def report(conf, **kwargs):
    """Builds an emulation stack report."""

    kwargs["as_json"] = kwargs.pop("json", False)
    wotemu.cli.report.build_report(conf, **kwargs)



import pandas as pd
import pytest
from wotemu.report.reader import explode_dict_column


@pytest.mark.asyncio
async def test_get_tasks(redis_reader, redis_test_data):
    tasks = await redis_reader.get_tasks()
    assert len(tasks) == redis_test_data.get_num_tasks()


@pytest.mark.asyncio
async def test_get_system_df(redis_reader, redis_test_data):
    task = redis_test_data.get_task_with_system_data()
    df = await redis_reader.get_system_df(task=task)

    assert set(df.index.names) == {"date"}

    columns = [
        "cpu_percent",
        "mem_mb",
        "mem_percent"
    ]

    for col in columns:
        assert df[col].notna().any()


@pytest.mark.asyncio
async def test_get_packet_df(redis_reader, redis_test_data):
    task = redis_test_data.get_task_with_packet_data()
    df = await redis_reader.get_packet_df(task=task)

    assert set(df.index.names) == {"date", "iface"}

    columns = [
        "len",
        "src",
        "dst",
        "proto",
        "transport",
        "srcport",
        "dstport"
    ]

    for col in columns:
        assert df[col].notna().any()


@pytest.mark.asyncio
async def test_get_info(redis_reader):
    tasks = await redis_reader.get_tasks()
    task = tasks.pop()
    infos = await redis_reader.get_info(task=task)

    assert len(infos) > 0

    info_item = infos[0]

    assert set(info_item.keys()) == {
        "boot_time",
        "cpu_count",
        "cpu_model",
        "env",
        "mem_total",
        "net",
        "process",
        "python_version",
        "time",
        "uname",
        "service_vips",
        "hostname",
        "container_id",
        "task_id",
        "constraints",
        "networks_cidr"
    }


@pytest.mark.asyncio
async def test_get_thing_df(redis_reader, redis_test_data):
    task = redis_test_data.get_task_with_thing_data()
    df = await redis_reader.get_thing_df(task=task)

    assert set(df.index.names) == {"date", "thing", "name", "verb"}

    columns = [
        "class",
        "host"
    ]

    for col in columns:
        assert df[col].notna().any()


@pytest.mark.asyncio
async def test_explode_dict_column():
    df = pd.DataFrame([
        {"key": "A0", "error": {"type": "ValueError", "message": "An error"}},
        {"key": "A1", "item": 10},
        {"key": "A2", "item": {"value": 20}}
    ])

    explode_dict_column(df, "item")

    assert df["item"].notna().any()
    assert df["item_value"].notna().any()

    explode_dict_column(df, "error")

    assert df["error"].notna().any()
    assert df["error_type"].notna().any()
    assert df["error_message"].notna().any()


@pytest.mark.asyncio
async def test_get_address_df(redis_reader):
    df = await redis_reader.get_address_df()

    assert set(df.index.names) == {"date", "iface", "task"}
    assert df["address"].notna().all()


@pytest.mark.asyncio
async def test_extend_packet_df(redis_reader, redis_test_data):
    task = redis_test_data.get_task_with_packet_data()
    df_packet = await redis_reader.get_packet_df(task=task)
    df = await redis_reader.extend_packet_df(df_packet)

    assert set(df.index.names) == {"date", "iface"}

    columns = [
        "len",
        "src",
        "dst",
        "proto",
        "transport",
        "srcport",
        "dstport",
        "src_task",
        "src_service",
        "dst_task",
        "dst_service"
    ]

    df_tcp = df[df["transport"] == "tcp"]
    assert df_tcp["src_service"].all()
    assert df_tcp["dst_service"].all()

    for col in columns:
        assert df[col].notna().any()


@pytest.mark.asyncio
async def test_get_service_traffic_df(redis_reader):
    df_in = await redis_reader.get_service_traffic_df(inbound=True)

    columns_in = [
        "len",
        "src_task",
        "dst_service"
    ]

    for col in columns_in:
        assert df_in[col].notna().any()

    df_out = await redis_reader.get_service_traffic_df(inbound=False)

    columns_out = [
        "len",
        "dst_task",
        "src_service"
    ]

    for col in columns_out:
        assert df_out[col].notna().any()


@pytest.mark.asyncio
async def test_get_snapshot_df(redis_reader):
    df = await redis_reader.get_snapshot_df()

    columns = [
        "desired_state",
        "is_running",
        "is_error",
        "task_id",
        "task",
        "service_id"
    ]

    for col in columns:
        assert df[col].notna().any()



import logging
import pprint

import docker
from wotemu.enums import Labels

_REDIS_PORT = 6379
_STACK_NAMESPACE = "com.docker.stack.namespace"

_logger = logging.getLogger(__name__)


def find_stack_services(stack, api_client=None):
    api_client = api_client or docker.APIClient()

    return [
        item for item in api_client.services()
        if item.get("Spec", {}).get("Labels", {}).get(_STACK_NAMESPACE) == stack
    ]


def find_stack_redis_port(stack, api_client=None):
    _logger.info("Finding publicly exposed Redis port for stack: %s", stack)

    services = find_stack_services(stack=stack, api_client=api_client)

    redis_service = next(
        item
        for item in services
        if item.get("Spec", {}).
        get("TaskTemplate", {}).
        get("ContainerSpec", {}).
        get("Labels", {}).
        get(Labels.WOTEMU_REDIS.value, None) is not None)

    _logger.debug("Found Redis service:\n%s", pprint.pformat(redis_service))

    return next(
        item["PublishedPort"]
        for item in redis_service["Endpoint"]["Ports"]
        if item["TargetPort"] == _REDIS_PORT and item["Protocol"] == "tcp")



import lxml.etree
from wotemu.report.utils import get_base_template


class BaseComponent:
    def to_element(self):
        raise NotImplementedError

    def to_page_html(self):
        tree, root = get_base_template()
        root.append(self.to_element())
        return lxml.etree.tostring(tree, method="html")



from datetime import datetime, timezone

import lxml.etree
from wotemu.report.components.base import BaseComponent
from wotemu.report.components.container import ContainerComponent
from wotemu.report.components.figure_block import FigureBlockComponent
from wotemu.report.components.figures_row import FiguresRowComponent


class TaskSectionComponent(BaseComponent):
    _LOGS_TITLE = "Logs"
    _LOGS_SUBTITLE = "Snapshot of the most recent log entries"
    _NOT_RUNNING = "The task was shut down prematurely (before the stack was manually stopped)"
    _ERROR = "It seems there was an error during the task execution"
    _INFO_TITLE = "Task details"
    _CONTAINER_ID = "Container ID"
    _NODE_ID = "Node ID"
    _SERVICE_ID = "Service ID"
    _TASK_ID = "Task ID"
    _CREATED_AT = "Created at"
    _UPDATED_AT = "Updated at"
    _STATUS = "Desired state"
    _STARTED_AT = "Started at"
    _LIM_MEM = "Memory limit (MB)"
    _LIM_CPU = "CPU limit (%)"

    def __init__(
            self, fig_mem, fig_cpu, fig_packet_iface, fig_packet_proto, fig_thing_counts,
            fig_cons_req_lat, fig_exps_req_lat, fig_cons_events, fig_exps_events, snapshot, info,
            title=None, height=450):
        self.fig_mem = fig_mem
        self.fig_cpu = fig_cpu
        self.fig_packet_iface = fig_packet_iface
        self.fig_packet_proto = fig_packet_proto
        self.fig_thing_counts = fig_thing_counts
        self.fig_cons_req_lat = fig_cons_req_lat
        self.fig_exps_req_lat = fig_exps_req_lat
        self.fig_cons_events = fig_cons_events
        self.fig_exps_events = fig_exps_events
        self.snapshot = snapshot
        self.info = info
        self.title = title
        self.height = height

    def _get_dt(self, text):
        dt = lxml.etree.Element("dt", attrib={"class": "col-sm-3 text-muted"})
        dt.text = text
        return dt

    def _get_dd(self, text, is_code=False):
        dd = lxml.etree.Element("dd", attrib={"class": "col-sm-9"})

        if is_code:
            code = lxml.etree.Element("code")
            code.text = text
            dd.append(code)
        else:
            dd.text = text

        return dd

    def _get_info_element(self):
        if not self.snapshot:
            return None

        card = lxml.etree.Element("div", attrib={"class": "card"})
        card_body = lxml.etree.Element("div", attrib={"class": "card-body"})
        card_title = lxml.etree.Element("h5", attrib={"class": "card-title"})
        card_title.text = self._INFO_TITLE
        card.append(card_body)
        card_body.append(card_title)

        dl = lxml.etree.Element("dl", attrib={"class": "row mb-0"})

        snap = self.snapshot
        info = self.info

        if snap.get("desired_state"):
            dl.append(self._get_dt(self._STATUS))
            dl.append(self._get_dd(snap["desired_state"]))

        if snap.get("node_id"):
            dl.append(self._get_dt(self._NODE_ID))
            dl.append(self._get_dd(snap["node_id"], is_code=True))

        if snap.get("service_id"):
            dl.append(self._get_dt(self._SERVICE_ID))
            dl.append(self._get_dd(snap["service_id"], is_code=True))

        if snap.get("task_id"):
            dl.append(self._get_dt(self._TASK_ID))
            dl.append(self._get_dd(snap["task_id"], is_code=True))

        if snap.get("container_id"):
            dl.append(self._get_dt(self._CONTAINER_ID))
            dl.append(self._get_dd(snap["container_id"], is_code=True))

        if snap.get("created_at"):
            dl.append(self._get_dt(self._CREATED_AT))
            dl.append(self._get_dd(snap["created_at"].isoformat()))

        if snap.get("updated_at"):
            dl.append(self._get_dt(self._UPDATED_AT))
            dl.append(self._get_dd(snap["updated_at"].isoformat()))

        if info.get("time"):
            start_dtime = datetime.fromtimestamp(info["time"], timezone.utc)
            dl.append(self._get_dt(self._STARTED_AT))
            dl.append(self._get_dd(start_dtime.isoformat()))

        if info.get("constraints", {}).get("cpu_percent"):
            dl.append(self._get_dt(self._LIM_CPU))
            dl.append(self._get_dd(str(info["constraints"]["cpu_percent"])))

        if info.get("constraints", {}).get("mem_limit_mb"):
            dl.append(self._get_dt(self._LIM_MEM))
            dl.append(self._get_dd(str(info["constraints"]["mem_limit_mb"])))

        card_body.append(dl)

        return card

    def _get_logs_element(self):
        if not self.snapshot or not self.snapshot.get("logs"):
            return None

        card = lxml.etree.Element("div", attrib={"class": "card"})
        is_error = self.snapshot.get("is_error", False)
        body_class = "bg-danger text-white" if is_error else "bg-light"
        body_class = f"card-body small {body_class}"
        card_body = lxml.etree.Element("div", attrib={"class": body_class})
        pre_class = "text-light" if is_error else ""
        pre_class = f"mb-0 {pre_class}"
        card_pre = lxml.etree.Element("pre", attrib={"class": pre_class})
        card_title = lxml.etree.Element("h5", attrib={"class": "card-title"})
        card_title.text = self._LOGS_TITLE
        sub_class = "" if is_error else "text-muted"
        sub_class = f"card-subtitle mb-2 {sub_class}"
        card_subtitle = lxml.etree.Element("h6", attrib={"class": sub_class})
        card_subtitle.text = self._LOGS_SUBTITLE
        card.append(card_body)
        card_body.append(card_title)
        card_body.append(card_subtitle)
        card_body.append(card_pre)
        card_pre.text = self.snapshot["logs"].strip()

        return card

    def _get_running_element(self):
        if not self.snapshot:
            return None

        if self.snapshot.get("is_running", True):
            return None

        alert = lxml.etree.Element(
            "div", attrib={"class": f"alert alert-warning"})

        alert.text = self._NOT_RUNNING

        return alert

    def _get_error_element(self):
        if not self.snapshot:
            return None

        if not self.snapshot.get("is_error", False):
            return None

        alert = lxml.etree.Element(
            "div", attrib={"class": f"alert alert-danger"})

        alert.text = self._ERROR

        return alert

    def _fig_el_ifdef(self, fig):
        if not fig:
            return None

        return FigureBlockComponent(fig).to_element()

    def to_element(self):
        figs = [
            self.fig_mem,
            self.fig_cpu,
            self.fig_packet_iface,
            self.fig_packet_proto
        ]

        figs = [item for item in figs if item is not None]

        figs_row = FiguresRowComponent(
            figs,
            col_class="col-xl-6",
            height=self.height).to_element()

        title = None

        if self.title:
            title = lxml.etree.Element("h4")
            title.text = self.title

        elements = [
            title,
            self._get_running_element(),
            self._get_error_element(),
            self._get_info_element(),
            figs_row,
            self._fig_el_ifdef(self.fig_thing_counts),
            self._fig_el_ifdef(self.fig_cons_req_lat),
            self._fig_el_ifdef(self.fig_exps_req_lat),
            self._fig_el_ifdef(self.fig_cons_events),
            self._fig_el_ifdef(self.fig_exps_events),
            self._get_logs_element()
        ]

        elements = [item for item in elements if item is not None]

        return ContainerComponent(elements=elements).to_element()



import copy
import os

from deepmerge import always_merger
from wotemu.config import DEFAULT_DOCKER_SOCKET, ConfigVars
from wotemu.enums import Labels

COMPOSE_VERSION = "3.7"
IMAGE_ENV_VAR = "WOTEMU_IMAGE_OVERRIDE"
BASE_IMAGE = "agmangas/wotemu"
TEMPLATE_TASK_NAME = "{{.Task.Name}}"
TEMPLATE_NODE_HOST = "{{.Node.Hostname}}"
TEMPLATE_NODE_ID = "{{.Node.ID}}"
TEMPLATE_SERVICE_ID = "{{.Service.ID}}"
TEMPLATE_SERVICE_NAME = "{{.Service.Name}}"
ENV_KEY_CPU_SPEED = "TARGET_CPU_SPEED"
ENV_KEY_NODE_HOST = "NODE_HOSTNAME"
ENV_KEY_NODE_ID = "NODE_ID"
ENV_KEY_SERVICE_ID = "SERVICE_ID"
ENV_KEY_SERVICE_NAME = "SERVICE_NAME"
ENV_KEY_CPU_SPEED = "TARGET_CPU_SPEED"
ENV_VAL_TRUTHY = "1"
VOL_DOCKER_SOCK = "{}:{}".format(DEFAULT_DOCKER_SOCKET, DEFAULT_DOCKER_SOCKET)

SERVICE_BASE_DOCKER_PROXY = {
    "image": "tecnativa/docker-socket-proxy",
    "environment": {
        "CONTAINERS": ENV_VAL_TRUTHY,
        "NETWORKS": ENV_VAL_TRUTHY,
        "TASKS": ENV_VAL_TRUTHY,
        "SERVICES": ENV_VAL_TRUTHY,
        "NODES": ENV_VAL_TRUTHY
    },
    "deploy": {
        "placement": {
            "constraints": ["node.role == manager"]
        }
    },
    "cap_add": ["ALL"],
    "volumes": [VOL_DOCKER_SOCK]
}

SERVICE_BASE_REDIS = {
    "image": "redis:5",
    "labels": {Labels.WOTEMU_REDIS.value: ""}
}

_ENVIRONMENT_BASE = {
    ENV_KEY_NODE_HOST: TEMPLATE_NODE_HOST,
    ENV_KEY_NODE_ID: TEMPLATE_NODE_ID,
    ENV_KEY_SERVICE_NAME: TEMPLATE_SERVICE_NAME,
    ENV_KEY_SERVICE_ID: TEMPLATE_SERVICE_ID
}

SERVICE_BASE_GENERIC = {
    "hostname": TEMPLATE_TASK_NAME,
    "labels": {Labels.WOTEMU_SERVICE.value: ""},
    "environment": {**_ENVIRONMENT_BASE}
}

SERVICE_BASE_GATEWAY = {
    "cap_add": ["ALL"],
    "hostname": TEMPLATE_TASK_NAME,
    "volumes": [VOL_DOCKER_SOCK],
    "labels": {Labels.WOTEMU_GATEWAY.value: ""},
    "environment": {**_ENVIRONMENT_BASE}
}

SERVICE_BASE_BROKER = {
    "cap_add": ["ALL"],
    "hostname": TEMPLATE_TASK_NAME,
    "volumes": [VOL_DOCKER_SOCK],
    "labels": {Labels.WOTEMU_BROKER.value: ""},
    "environment": {**_ENVIRONMENT_BASE}
}

SERVICE_BASE_NODE = {
    "cap_add": ["ALL"],
    "hostname": TEMPLATE_TASK_NAME,
    "volumes": [VOL_DOCKER_SOCK],
    "labels": {Labels.WOTEMU_NODE.value: ""},
    "environment": {**_ENVIRONMENT_BASE}
}

NETWORK_BASE = {
    "driver": "overlay",
    "attachable": True,
    "labels": {Labels.WOTEMU_NETWORK.value: ""}
}

NETWORK_SERVICE_BASE = {
    "driver": "overlay",
    "attachable": True,
    "labels": {Labels.WOTEMU_SERVICE_NETWORK.value: ""}
}


def _merge_topology_config(service, topology):
    envr = service.get("environment", {})

    envr.update({
        key: str(val)
        for key, val in topology.config.items()
        if val is not None
    })

    service["environment"] = envr


def get_docker_proxy_definition(topology):
    service = copy.deepcopy(SERVICE_BASE_DOCKER_PROXY)

    service.update({
        "networks": [net.name for net in topology.networks]
    })

    return {topology.docker_proxy.host: service}


def get_redis_definition(topology, topology_redis):
    service = copy.deepcopy(SERVICE_BASE_REDIS)

    service.update({
        "networks": [net.name for net in topology.networks],
        "ports": ["6379"]
    })

    return {topology.redis.host: service}


def get_generic_service_definition(topology, srv):
    service = copy.deepcopy(SERVICE_BASE_GENERIC)

    if srv.params.get("env"):
        service["environment"].update(srv.params["env"])

    if srv.params.get("labels"):
        service["labels"].update(srv.params["labels"])

    deploy = {}

    if srv.params.get("restart_policy"):
        policy = srv.params["restart_policy"]

        deploy.update({
            "restart_policy": {
                "condition": policy["Condition"]
            }
        })

    if srv.params.get("mode"):
        mode = srv.params["mode"]
        deploy.update({"mode": mode["Mode"]})

    if len(deploy) > 0:
        service.update({"deploy": deploy})

    service.update({
        "image": srv.image,
        "networks": [
            srv.get_node_network_name(node)
            for node in topology.get_service_nodes(srv)
        ]
    })

    return {srv.name: service}


def get_network_gateway_definition(topology, network):
    service = copy.deepcopy(SERVICE_BASE_GATEWAY)

    depends_on = []

    if topology.docker_proxy.enabled:
        depends_on.append(topology.docker_proxy.host)

    if topology.redis.enabled:
        depends_on.append(topology.redis.host)

    service.update({
        "image": os.getenv(IMAGE_ENV_VAR, BASE_IMAGE),
        "command": network.cmd_gateway,
        "networks": [network.name],
        "depends_on": depends_on
    })

    if network.args_compose_gw:
        always_merger.merge(service, network.args_compose_gw)

    _merge_topology_config(service, topology)

    return {network.name_gateway: service}


def get_network_definition(topology, network):
    definition = copy.deepcopy(NETWORK_BASE)
    definition.update({"name": network.name})
    return {network.name: definition}


def get_broker_definition(topology, broker):
    service = copy.deepcopy(SERVICE_BASE_BROKER)

    depends_on = [
        *[net.name_gateway for net in broker.networks]
    ]

    if topology.docker_proxy.enabled:
        depends_on.append(topology.docker_proxy.host)

    if topology.redis.enabled:
        depends_on.append(topology.redis.host)

    service.update({
        "image": os.getenv(IMAGE_ENV_VAR, BASE_IMAGE),
        "command": broker.cmd,
        "networks": [net.name for net in broker.networks],
        "depends_on": depends_on
    })

    if broker.args_compose:
        always_merger.merge(service, broker.args_compose)

    _merge_topology_config(service, topology)

    return {broker.name: service}


def get_node_resources_definition(node_resources):
    limits = {}

    if node_resources.cpu_limit:
        limits.update({"cpus": node_resources.cpu_limit})

    if node_resources.mem_limit:
        limits.update({"memory": node_resources.mem_limit})

    reserv = {}

    if node_resources.cpu_reservation:
        reserv.update({"cpus": node_resources.cpu_reservation})

    if node_resources.mem_reservation:
        reserv.update({"memory": node_resources.mem_reservation})

    ret = {}

    if len(limits) > 0:
        ret.update({"limits": limits})

    if len(reserv) > 0:
        ret.update({"reservations": reserv})

    return ret if len(ret) > 0 else None


def get_node_definition(topology, node):
    service = copy.deepcopy(SERVICE_BASE_NODE)

    networks = [net.name for net in node.networks]

    for srv in node.services:
        networks.append(srv.get_node_network_name(node))

    depends_on = [
        *[net.name_gateway for net in node.networks]
    ]

    if topology.docker_proxy.enabled:
        depends_on.append(topology.docker_proxy.host)

    if topology.redis.enabled:
        depends_on.append(topology.redis.host)

    envr = service.get("environment", {})

    if node.broker:
        assert node.broker_network
        networks.append(node.broker_network.name)
        depends_on.append(node.broker_network.name_gateway)
        envr.update({ConfigVars.MQTT_BROKER_HOST.value: node.broker_host})

    service.update({
        "image": node.image,
        "command": node.cmd_app,
        "networks": list(set(networks)),
        "depends_on": list(set(depends_on)),
        "environment": envr
    })

    deploy = service.get("deploy", {})

    if node.resources:
        resources_dict = get_node_resources_definition(node.resources)

        if resources_dict:
            deploy.update({"resources": resources_dict})

        if node.resources.target_cpu_speed:
            env_speed = {ENV_KEY_CPU_SPEED: node.resources.target_cpu_speed}
            always_merger.merge(service, {"environment": env_speed})

    if node.scale:
        deploy.update({"replicas": node.scale})

    if len(deploy) > 0:
        service.update({"deploy": deploy})

    if node.args_compose:
        always_merger.merge(service, node.args_compose)

    _merge_topology_config(service, topology)

    return {node.name: service}


def _service_networks(topology, srv):
    nets = {}

    for node in topology.get_service_nodes(srv):
        net_name = srv.get_node_network_name(node)
        definition = copy.deepcopy(NETWORK_SERVICE_BASE)
        definition.update({"name": net_name})
        nets.update({net_name: definition})

    return nets


def get_topology_definition(topology):
    definition = {"version": COMPOSE_VERSION}

    services = {}

    if topology.docker_proxy.enabled:
        services.update(topology.docker_proxy_compose_dict)

    if topology.redis.enabled:
        services.update(topology.redis_compose_dict)

    for net in topology.networks:
        services.update(net.to_gateway_compose_dict(topology))

    for srv in topology.services:
        services.update(srv.to_compose_dict(topology))

    for node in topology.nodes:
        services.update(node.to_compose_dict(topology))

    for broker in topology.brokers:
        services.update(broker.to_compose_dict(topology))

    networks = {}

    for net in topology.networks:
        networks.update(net.to_compose_dict(topology))

    for srv in topology.services:
        networks.update(_service_networks(topology, srv))

    definition.update({
        "services": services,
        "networks": networks
    })

    return definition



import functools
import json
import random

import pytest
from tests.conftest import TD_EXAMPLE
from wotpy.wot.td import ThingDescription

from wotemu.wotpy.redis import redis_thing_callback
from wotemu.wotpy.wot import wot_entrypoint

WOT_HOSTNAME = "127.0.0.1"


@pytest.mark.asyncio
async def test_redis_callback(redis, unused_tcp_port_factory):
    redis_cb = functools.partial(redis_thing_callback, client=redis)

    wot = wot_entrypoint(
        port_catalogue=unused_tcp_port_factory(),
        hostname=WOT_HOSTNAME,
        exposed_cb=redis_cb,
        consumed_cb=redis_cb,
        port_http=unused_tcp_port_factory())

    exposed_thing = wot.produce(model=json.dumps(TD_EXAMPLE))
    exposed_thing.expose()
    await exposed_thing.servient.start()

    td_str = ThingDescription.from_thing(exposed_thing.thing).to_str()
    consumed_thing = wot.consume(td_str)

    _cur, keys = await redis.scan(b"0")
    assert len(keys) == 0

    val = int(random.random() * 100)
    await consumed_thing.properties["testProp"].write(val)
    read_val = await consumed_thing.properties["testProp"].read()
    assert read_val == val

    _cur, keys = await redis.scan(b"0")
    assert len(keys) > 0



"""
A MongoDB-based historian Thing able to persist property 
values and return them in an aggregated fashion.
"""

import asyncio
import collections
import functools
import json
import logging
import os
import pprint
import time
from datetime import datetime

import motor.motor_asyncio
import pymongo
from wotemu.utils import consume_from_catalogue
from wotpy.wot.td import ThingDescription

_DEFAULT_FALLBACK_DB = "wotemu_mongo_historian"
_DEFAULT_QUERY_SECONDS = 60

_DESCRIPTION = {
    "id": "urn:org:fundacionctic:thing:historian",
    "name": "MongoDB historian Thing",
    "properties": {
        "observedThings": {
            "readOnly": True,
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "servient": {
                        "type": "string"
                    },
                    "thingId": {
                        "type": "string"
                    }
                },
                "required": ["servient", "thingId"]
            }
        }
    },
    "actions": {
        "list": {
            "safe": True,
            "idempotent": False,
            "output": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "thingId": {
                            "type": "string"
                        },
                        "propertyName": {
                            "type": "string"
                        },
                        "servient": {
                            "type": "string"
                        }
                    },
                    "required": ["propertyName", "thingId", "servient"]
                }
            }
        },
        "get": {
            "safe": True,
            "idempotent": False,
            "input": {
                "type": "object",
                "properties": {
                    "fromUnix": {
                        "type": "number"
                    },
                    "toUnix": {
                        "type": "number"
                    },
                    "thingId": {
                        "type": "string"
                    },
                    "propertyName": {
                        "type": "string"
                    },
                    "servient": {
                        "type": "string"
                    },
                    "buckets": {
                        "type": "number"
                    }
                }
            },
            "output": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "unixTime": {
                            "type": "number"
                        },
                        "thingId": {
                            "type": "string"
                        },
                        "propertyName": {
                            "type": "string"
                        },
                        "servient": {
                            "type": "string"
                        },
                        "value": {
                            "type": "string"
                        }
                    },
                    "required": ["unixTime", "propertyName", "thingId", "servient", "value"]
                }
            }
        },
        "write": {
            "safe": False,
            "idempotent": False,
            "input": {
                "type": "object",
                "properties": {
                    "unixTime": {
                        "type": "number"
                    },
                    "propertyName": {
                        "type": "string"
                    },
                    "thingId": {
                        "type": "string"
                    },
                    "servient": {
                        "type": "string"
                    },
                    "value": {
                        "type": "string"
                    }
                },
                "required": ["propertyName", "thingId", "servient", "value"]
            }
        }
    }
}

_logger = logging.getLogger(__name__)


MongoHistorianThing = collections.namedtuple(
    "MongoHistorianThing",
    ["task_observe", "exposed_thing", "observe_stop_event"])


async def _get_aggregated(params_input, motor_client, db_name, query):
    params_input = params_input or {}
    num_buckets = int(params_input.get("buckets", 10))

    pipeline = [
        {
            "$match": query
        },
        {
            "$addFields": {
                "date_str": {
                    "$dateToString": {
                        "format": "%Y%m%d_%H%M%S",
                        "date": "$utc_date"
                    }
                }
            }
        },
        {
            "$bucketAuto": {
                "groupBy": "$date_str",
                "buckets": num_buckets,
                "output": {
                    "utc_date": {"$first": "$utc_date"},
                    "property_name": {"$first": "$property_name"},
                    "thing_id": {"$first": "$thing_id"},
                    "servient": {"$first": "$servient"},
                    "value": {"$first": "$value"}
                }
            }
        },
        {
            "$sort": {"utc_date": 1}
        }
    ]

    _logger.debug("Running pipeline:\n%s", pprint.pformat(pipeline))

    cur = motor_client[db_name].properties.aggregate(pipeline)

    return [
        {
            "thingId": doc["thing_id"],
            "propertyName": doc["property_name"],
            "servient": doc["servient"],
            "value": doc["value"],
            "unixTime": doc["utc_date"].timestamp()
        }
        async for doc in cur
    ]


async def _get(params, motor_client, db_name):
    params_input = params["input"]
    params_input = params_input or {}

    from_unix_default = time.time() - _DEFAULT_QUERY_SECONDS
    from_unix = int(params_input.get("fromUnix", from_unix_default))

    query = {
        "utc_date": {
            "$gte": datetime.utcfromtimestamp(from_unix)
        }
    }

    if params_input.get("toUnix"):
        to_unix = int(params_input["toUnix"])
        query["utc_date"].update({"$lt": datetime.utcfromtimestamp(to_unix)})

    if params_input.get("propertyName"):
        query.update({
            "property_name": {
                "$eq": params_input["propertyName"]
            }
        })

    if params_input.get("thingId"):
        query.update({
            "thing_id": {
                "$eq": params_input["thingId"]
            }
        })

    if params_input.get("servient"):
        query.update({
            "servient": {
                "$eq": params_input["servient"]
            }
        })

    if params_input.get("buckets"):
        return await _get_aggregated(
            params_input=params_input,
            motor_client=motor_client,
            db_name=db_name,
            query=query)

    cur = motor_client[db_name].properties.find(query).sort("utc_date")

    return [
        {
            "thingId": doc["thing_id"],
            "propertyName": doc["property_name"],
            "servient": doc["servient"],
            "value": doc["value"],
            "unixTime": doc["utc_date"].timestamp()
        }
        async for doc in cur
    ]


async def _write(params, motor_client, db_name):
    params_input = params["input"]
    params_input = params_input or {}

    tstamp = params_input.get("unixTime")
    dtime = datetime.utcfromtimestamp(tstamp) if tstamp else datetime.utcnow()

    doc = {
        "utc_date": dtime,
        "property_name": params_input["propertyName"],
        "thing_id": params_input["thingId"],
        "servient": params_input["servient"],
        "value": params_input["value"]
    }

    _logger.debug("Inserting doc: %s", doc)
    result = await motor_client[db_name].properties.insert_one(doc)
    _logger.debug("Inserted: %s", result.inserted_id)


async def _list(params, motor_client, db_name):
    pipeline = [
        {
            "$group": {
                "_id": {
                    "property_name": "$property_name",
                    "thing_id": "$thing_id",
                    "servient": "$servient"
                }
            }
        }
    ]

    _logger.debug("Running pipeline:\n%s", pprint.pformat(pipeline))

    cur = motor_client[db_name].properties.aggregate(pipeline)

    return [
        {
            "propertyName": doc["_id"]["property_name"],
            "thingId": doc["_id"]["thing_id"],
            "servient": doc["_id"]["servient"]
        }
        async for doc in cur
    ]


async def _wait_mongo(mongo_uri, timeout_ping, sleep_iter=1.0):
    now = time.time()
    time_limit = now + timeout_ping

    _logger.info("Waiting for MongoDB (%s secs)", timeout_ping)

    while time.time() < time_limit:
        try:
            motor_client = motor.motor_asyncio.AsyncIOMotorClient(mongo_uri)
            assert await motor_client.admin.command("ping")
            return True
        except Exception as ex:
            _logger.debug("Still waiting for MongoDB: %s", repr(ex))

        await asyncio.sleep(sleep_iter)

    raise RuntimeError("Timeout waiting for MongoDB")


async def _create_indexes(motor_client, db_name):
    keys = [
        ("utc_date", pymongo.DESCENDING),
        ("property_name", pymongo.DESCENDING),
        ("thing_id", pymongo.DESCENDING),
        ("servient", pymongo.DESCENDING)
    ]

    _logger.info("Creating index: %s", keys)

    await motor_client[db_name].properties.create_index(
        keys,
        name="utc_date_property_name_thing_id_servient",
        background=True)

    keys = [
        ("property_name", pymongo.DESCENDING),
        ("thing_id", pymongo.DESCENDING),
        ("servient", pymongo.DESCENDING)
    ]

    _logger.info("Creating index: %s", keys)

    await motor_client[db_name].properties.create_index(
        keys,
        name="property_name_thing_id_servient",
        background=True)


async def _run_observe_historian_iter(
        wot, port_catalogue, servient_host, thing_id,
        buckets, interval, motor_client, db_name, invoke_timeout):
    to_unix = datetime.utcnow().timestamp()
    from_unix = to_unix - interval

    _logger.debug(
        "Running historian observation iteration (from=%s) (to=%s)",
        datetime.utcfromtimestamp(from_unix),
        datetime.utcfromtimestamp(to_unix))

    cons_thing = await consume_from_catalogue(
        wot=wot,
        port_catalogue=port_catalogue,
        servient_host=servient_host,
        thing_id=thing_id)

    props_list = await cons_thing.actions["list"].invoke(None, timeout=invoke_timeout)

    _logger.debug("Reading properties:\n%s", pprint.pformat(props_list))

    invocation_params = [
        {
            "toUnix": to_unix,
            "fromUnix": from_unix,
            "buckets": buckets,
            "propertyName": prop_item["propertyName"],
            "thingId": prop_item["thingId"],
            "servient": prop_item["servient"]
        }
        for prop_item in props_list
    ]

    invocation_awaitables = [
        cons_thing.actions["get"].invoke(params, timeout=invoke_timeout)
        for params in invocation_params
    ]

    results = await asyncio.gather(*invocation_awaitables, return_exceptions=True)

    write_params = [
        {
            "input": {
                "unixTime": item["unixTime"],
                "propertyName": item["propertyName"],
                "thingId": item["thingId"],
                "servient": item["servient"],
                "value": item["value"]
            }
        }
        for results_list in results
        for item in results_list
        if not isinstance(results_list, Exception)
    ]

    write_awaitables = [
        _write(params=params, motor_client=motor_client, db_name=db_name)
        for params in write_params
    ]

    await asyncio.gather(*write_awaitables, return_exceptions=True)


async def _start_observe_historian_task(
        wot, port_catalogue, servient_host, thing_id,
        buckets, interval, motor_client, db_name, stop_event, invoke_timeout=40):
    run_iter = functools.partial(
        _run_observe_historian_iter,
        wot=wot,
        port_catalogue=port_catalogue,
        servient_host=servient_host,
        thing_id=thing_id,
        buckets=buckets,
        interval=interval,
        motor_client=motor_client,
        db_name=db_name,
        invoke_timeout=invoke_timeout)

    prev_time = None
    next_time = time.time() + interval

    while not stop_event.is_set():
        now = time.time()

        if now >= next_time:
            if prev_time:
                delta = now - prev_time
                _logger.debug("Historian observation task delta: %s s.", delta)

            prev_time = now
            next_time = now + interval

            try:
                await run_iter()
            except Exception as ex:
                _logger.warning("Error in historian iteration: %s", repr(ex))

        await asyncio.sleep(1)


async def _read_insert_property(consumed_thing, name, thing_id, servient_host, motor_client, db_name):
    try:
        _logger.debug("Reading property (%s)", name)
        val = await consumed_thing.properties[name].read()
        _logger.debug("Read property (%s=%s)", name, val)
    except Exception as ex:
        _logger.warning("Error reading (%s): %s", name, repr(ex))
        return

    params = {
        "input": {
            "propertyName": name,
            "thingId": thing_id,
            "servient": servient_host,
            "value": val
        }
    }

    await _write(params=params, motor_client=motor_client, db_name=db_name)


async def _run_observe_thing_iter(wot, port_catalogue, motor_client, db_name, servient_host, thing_id):
    cons_thing = await consume_from_catalogue(
        wot=wot,
        port_catalogue=port_catalogue,
        servient_host=servient_host,
        thing_id=thing_id)

    read_insert_kwargs = [
        {
            "consumed_thing": cons_thing,
            "name": name,
            "thing_id": thing_id,
            "servient_host": servient_host,
            "motor_client": motor_client,
            "db_name": db_name
        }
        for name in cons_thing.properties
    ]

    read_insert_awaitables = [
        _read_insert_property(**item)
        for item in read_insert_kwargs
    ]

    await asyncio.gather(*read_insert_awaitables, return_exceptions=True)


async def _start_observe_thing_task(wot, port_catalogue, motor_client, db_name, observed_things, interval, stop_event):
    run_iter = functools.partial(
        _run_observe_thing_iter,
        wot=wot,
        port_catalogue=port_catalogue,
        motor_client=motor_client,
        db_name=db_name)

    while not stop_event.is_set():
        iter_awaitables = [
            run_iter(
                servient_host=item["servient_host"],
                thing_id=item["thing_id"])
            for item in observed_things
        ]

        await asyncio.gather(*iter_awaitables, return_exceptions=True)
        await asyncio.sleep(interval)


async def _start_observe_task(
        wot, port_catalogue, observed_things, motor_client, db_name, interval, stop_event,
        downlink_servient_host, downlink_thing_id, downlink_buckets, downlink_interval):
    awaitables = []

    if observed_things and len(observed_things) > 0:
        thing_awaitable = _start_observe_thing_task(
            wot=wot,
            port_catalogue=port_catalogue,
            motor_client=motor_client,
            db_name=db_name,
            observed_things=observed_things,
            interval=interval,
            stop_event=stop_event)

        awaitables.append(thing_awaitable)

    if downlink_servient_host and downlink_thing_id:
        historian_awaitable = _start_observe_historian_task(
            wot=wot,
            port_catalogue=port_catalogue,
            servient_host=downlink_servient_host,
            thing_id=downlink_thing_id,
            buckets=downlink_buckets,
            interval=downlink_interval,
            motor_client=motor_client,
            db_name=db_name,
            stop_event=stop_event)

        awaitables.append(historian_awaitable)

    await asyncio.gather(*awaitables, return_exceptions=False)


async def _things_handler(observed_things):
    return [
        {
            "servient": item["servient_host"],
            "thingId": item["thing_id"]
        }
        for item in observed_things
    ]


async def produce_thing(
        wot, conf, mongo_uri, db_name=None, observed_things=None, interval=5, timeout_ping=60,
        downlink_servient_host=None, downlink_thing_id=None, downlink_buckets=1, downlink_interval=60):
    downlink_args = [downlink_servient_host, downlink_thing_id]

    if any(downlink_args) and not all(downlink_args):
        raise ValueError("Partially defined downlink parameters")

    downlink_buckets = int(downlink_buckets)
    downlink_interval = int(downlink_interval)

    await _wait_mongo(mongo_uri=mongo_uri, timeout_ping=timeout_ping)

    _logger.info(
        "Producing Thing:\n%s",
        pprint.pformat(_DESCRIPTION))

    _logger.info("Connecting to MongoDB: %s", mongo_uri)

    motor_client = motor.motor_asyncio.AsyncIOMotorClient(mongo_uri)
    assert await motor_client.admin.command("ping")
    db_name = db_name or os.getenv("SERVICE_NAME", _DEFAULT_FALLBACK_DB)

    _logger.info("MongoDB connection OK (db: %s)", db_name)

    await _create_indexes(motor_client=motor_client, db_name=db_name)

    handler_get = functools.partial(
        _get,
        motor_client=motor_client,
        db_name=db_name)

    handler_write = functools.partial(
        _write,
        motor_client=motor_client,
        db_name=db_name)

    handler_list = functools.partial(
        _list,
        motor_client=motor_client,
        db_name=db_name)

    try:
        observed_things = observed_things or []
        observed_things = json.loads(observed_things)
    except:
        pass

    things_handler = functools.partial(
        _things_handler,
        observed_things=observed_things)

    exposed_thing = wot.produce(json.dumps(_DESCRIPTION))
    exposed_thing.set_action_handler("get", handler_get)
    exposed_thing.set_action_handler("write", handler_write)
    exposed_thing.set_action_handler("list", handler_list)
    exposed_thing.set_property_read_handler("observedThings", things_handler)

    stop_event = asyncio.Event()

    task_observe = asyncio.ensure_future(_start_observe_task(
        wot=wot,
        port_catalogue=conf.port_catalogue,
        observed_things=observed_things,
        motor_client=motor_client,
        db_name=db_name,
        interval=interval,
        stop_event=stop_event,
        downlink_servient_host=downlink_servient_host,
        downlink_thing_id=downlink_thing_id,
        downlink_buckets=downlink_buckets,
        downlink_interval=downlink_interval))

    return MongoHistorianThing(
        task_observe=task_observe,
        exposed_thing=exposed_thing,
        observe_stop_event=stop_event)


async def app(wot, conf, loop, *args, **kwargs):
    historian_thing = await produce_thing(wot=wot, conf=conf, *args, **kwargs)
    historian_thing.exposed_thing.expose()

    td = ThingDescription.from_thing(historian_thing.exposed_thing.thing)
    _logger.debug("Exposed Thing:\n%s", pprint.pformat(td.to_dict()))

    try:
        await historian_thing.task_observe
    except asyncio.CancelledError:
        historian_thing.observe_stop_event.set()
        await asyncio.wait_for(historian_thing.task_observe, timeout=10)



from datetime import datetime

import lxml.etree
from wotemu.report.components.base import BaseComponent
from wotemu.report.components.container import ContainerComponent
from wotemu.report.components.figure_block import FigureBlockComponent


class TaskListComponent(BaseComponent):
    DEFAULT_TITLE = "List of tasks in the emulation stack"

    def __init__(self, task_keys, task_infos, df_snapshot, title=DEFAULT_TITLE, to_href=None):
        self.task_keys = task_keys
        self.task_infos = task_infos
        self.df_snapshot = df_snapshot
        self.title = title
        self.to_href = to_href if to_href else lambda x: f"{x}.html"

    def _get_item_class(self, task_key):
        df_snap = self.df_snapshot[self.df_snapshot["task"] == task_key]
        ret = "list-group-item list-group-item-action"

        if not df_snap.empty and df_snap["is_error"].any():
            ret = f"{ret} list-group-item-danger"
        elif not df_snap.empty and not df_snap["is_running"].all():
            ret = f"{ret} list-group-item-warning"
        else:
            ret = f"{ret} text-primary"

        return ret

    def _get_init_dtime(self, task_key):
        boot_time = self.task_infos.get(task_key, {}).get("time")

        if not boot_time:
            return None

        return datetime.utcfromtimestamp(boot_time)

    def _get_sorted_task_keys(self):
        def key(task):
            dtime = self._get_init_dtime(task)
            return dtime.timestamp() if dtime else 0

        return sorted(self.task_keys, key=key)

    def _get_item_element(self, task_key):
        item = lxml.etree.Element("a", attrib={
            "class": self._get_item_class(task_key),
            "href": self.to_href(task_key)
        })

        span = lxml.etree.Element("span")
        span.text = task_key
        item.append(span)

        dtime = self._get_init_dtime(task_key)

        if not dtime:
            return item

        dtime_span = lxml.etree.Element(
            "span", attrib={"class": "text-muted small"})

        dtime_span.text = dtime.isoformat()
        item.append(lxml.etree.Element("br"))
        item.append(dtime_span)

        return item

    def to_element(self):
        task_links = [
            self._get_item_element(task_key)
            for task_key in self._get_sorted_task_keys()
        ]

        title = lxml.etree.Element("h4")
        title_main = lxml.etree.Element("span")
        title_main.text = self.title
        sm_class = "ml-2 text-muted"
        title_small = lxml.etree.Element("small", attrib={"class": sm_class})
        title_small.text = "Sorted by start time"
        title.append(title_main)
        title.append(title_small)

        list_group = lxml.etree.Element("div", attrib={"class": "list-group"})
        [list_group.append(item) for item in task_links]

        return ContainerComponent(elements=[title, list_group]).to_element()



import tempfile
import xml.etree.ElementTree as ET

import html5lib
import pytest
from wotemu.report.builder import ReportBuilder


@pytest.mark.asyncio
async def test_build_task_mem_figure(redis_reader, redis_test_data):
    builder = ReportBuilder(reader=redis_reader)
    task = redis_test_data.get_task_with_system_data()
    fig = await builder.build_task_mem_figure(task=task)
    assert fig


@pytest.mark.asyncio
async def test_build_report(redis_reader):
    builder = ReportBuilder(reader=redis_reader)
    report_pages = await builder.build_report()

    for file_bytes in report_pages.values():
        with tempfile.TemporaryFile() as fp:
            fp.write(file_bytes)
            fp.seek(0)
            bytes_read = fp.read()
            document = html5lib.parse(bytes_read.decode())
            assert document


@pytest.mark.asyncio
async def test_write_report(redis_reader):
    builder = ReportBuilder(reader=redis_reader)

    with tempfile.TemporaryDirectory() as tmp_dir:
        await builder.write_report(base_path=tmp_dir)


@pytest.mark.asyncio
async def test_write_report_dataset(redis_reader):
    builder = ReportBuilder(reader=redis_reader)

    with tempfile.TemporaryDirectory() as tmp_dir:
        await builder.write_report_dataset(base_path=tmp_dir)



import lxml.etree
from wotemu.report.components.base import BaseComponent
from wotemu.report.components.figure_block import FigureBlockComponent


class FiguresRowComponent(BaseComponent):
    def __init__(self, figs, title=None, height=450, col_class="col-xl"):
        self.figs = figs
        self.title = title
        self.height = height
        self.col_class = col_class

    def to_element(self):
        fig_components = [
            FigureBlockComponent(
                fig,
                title=None,
                height=self.height,
                col_class=self.col_class,
                with_row=False)
            for fig in self.figs
        ]

        el_row = lxml.etree.Element("div", attrib={"class": "row"})

        if self.title:
            el_title = lxml.etree.Element("h4")
            el_title.text = self.title

            el_title_col = lxml.etree.Element(
                "div", attrib={"class": "col-xl-12"})

            el_title_col.append(el_title)
            el_row.append(el_title_col)

        [el_row.append(item.to_element()) for item in fig_components]

        return el_row



import asyncio
import functools
import json
import logging
import pprint
import socket
import time

import aioredis
import wotemu.config
from wotemu.enums import RedisPrefixes
from wotemu.monitor.packet import monitor_packets
from wotemu.monitor.system import get_node_info, monitor_system

_logger = logging.getLogger(__name__)


class NodeMonitor:
    def __init__(
            self, key=None, redis_url=None, packet_ifaces=None,
            packet_kwargs=None, system_kwargs=None):
        conf = wotemu.config.get_env_config()
        self._conf = conf
        self._key = key if key else socket.getfqdn()
        self._redis_url = redis_url if redis_url else conf.redis_url
        self._redis = None
        self._packet_ifaces = packet_ifaces
        self._packet_kwargs = packet_kwargs if packet_kwargs else {}
        self._system_kwargs = system_kwargs if system_kwargs else {}
        self._task_system = None
        self._tasks_packet = None

    @property
    def is_running(self):
        return self._task_system or self._tasks_packet

    async def _redis_open(self):
        if self._redis:
            return

        self._redis = await aioredis.create_redis_pool(self._redis_url)
        _logger.debug("Opened Redis connection to: %s", self._redis_url)

    async def _redis_close(self):
        if not self._redis:
            return

        self._redis.close()
        await self._redis.wait_closed()
        self._redis = None
        _logger.debug("Closed Redis connection")

    async def _redis_callback(self, items, key):
        _logger.debug("ZADD (%s items): %s", len(items), key)

        tr = self._redis.multi_exec()

        for item in items:
            score = item.get("time", time.time())
            member = json.dumps(item)
            tr.zadd(key=key, score=score, member=member)

        exec_res = await tr.execute()

        if not all(exec_res):
            _logger.warning("Error in Redis MULTI ZADD: %s", exec_res)

    async def _create_system_task(self):
        assert not self._task_system

        key = "{}:{}:{}".format(
            RedisPrefixes.NAMESPACE.value,
            RedisPrefixes.SYSTEM.value,
            self._key)

        async_cb = functools.partial(self._redis_callback, key=key)

        system_awaitable = monitor_system(
            async_cb=async_cb,
            **self._system_kwargs)

        self._task_system = asyncio.ensure_future(system_awaitable)

    async def _create_packet_tasks(self):
        assert not self._tasks_packet

        if not self._packet_ifaces or not len(self._packet_ifaces):
            _logger.warning("No packet interfaces were defined")
            return

        packet_awaitables = []

        for iface in self._packet_ifaces:
            key = "{}:{}:{}:{}".format(
                RedisPrefixes.NAMESPACE.value,
                RedisPrefixes.PACKET.value,
                iface,
                self._key)

            async_cb = functools.partial(self._redis_callback, key=key)

            packet_awaitables.append(monitor_packets(
                conf=self._conf,
                interface=iface,
                async_cb=async_cb))

        self._tasks_packet = [
            asyncio.ensure_future(item)
            for item in packet_awaitables
        ]

    async def _stop_system_task(self):
        if not self._task_system:
            return

        self._task_system.cancel()
        await self._task_system
        self._task_system = None

    async def _stop_packet_tasks(self):
        if not self._tasks_packet:
            return

        [task.cancel() for task in self._tasks_packet]
        await asyncio.gather(*self._tasks_packet)
        self._tasks_packet = None

    async def _write_node_info(self):
        assert self._redis

        node_info = get_node_info()
        tstamp = time.time()
        node_info.update({"time": tstamp})
        member = json.dumps(node_info)

        key = "{}:{}:{}".format(
            RedisPrefixes.NAMESPACE.value,
            RedisPrefixes.INFO.value,
            self._key)

        _logger.debug(
            "Writing node info - ZADD %s:\n%s",
            key, pprint.pformat(node_info))

        await self._redis.zadd(key=key, score=tstamp, member=member)

    async def start(self):
        if self.is_running:
            raise RuntimeError("Already running")

        _logger.debug("Starting node monitor")

        await self._redis_open()
        await self._write_node_info()
        await self._create_system_task()
        await self._create_packet_tasks()

    async def stop(self):
        _logger.debug("Stopping node monitor")

        await self._stop_system_task()
        await self._stop_packet_tasks()
        await self._redis_close()

        self._task_system = None
        self._tasks_packet = None



import asyncio
import functools
import logging
import pprint
import signal

import sh
from wotemu.monitor.base import NodeMonitor
from wotemu.utils import strip_ansi_codes

_ERR_MOSQUITTO = "error"

_logger = logging.getLogger(__name__)


def _loop_ex_handler(loop, context):
    _logger.warning("Exception in loop:\n%s", pprint.pformat(context))


def _done(cmd, success, exit_code, loop, close_loop):
    _logger.log(
        logging.INFO if success else logging.WARNING,
        "[%s] Exit code: %s",
        cmd.ran,
        exit_code)

    if not success:
        _logger.warning("Mosquitto terminated with error: %s", exit_code)
        asyncio.run_coroutine_threadsafe(close_loop(), loop)


def _out(line):
    line = strip_ansi_codes(line.strip())
    level = logging.WARNING if _ERR_MOSQUITTO in line.lower() else logging.INFO
    _logger.log(level, "[mosquitto] %s", line)


def _run_broker(conf, loop, close_loop):
    sh_mosquitto = sh.Command("mosquitto")

    cmd = [
        "-p",
        f"{conf.port_mqtt}"
    ]

    _logger.info("Running mosquitto: %s", cmd)

    done = functools.partial(
        _done,
        loop=loop,
        close_loop=close_loop)

    proc = sh_mosquitto(
        cmd,
        _err_to_out=True,
        _out=_out,
        _bg=True,
        _done=done)

    return proc


def _terminate_broker(proc):
    try:
        _logger.info("Terminating broker")
        proc.terminate()
        proc.wait()
    except:
        _logger.warning("Error terminating broker", exc_info=True)


async def _close_loop(loop, monitor):
    if monitor:
        try:
            _logger.debug("Stopping monitor")
            await monitor.stop()
        except:
            _logger.warning("Error stopping monitor", exc_info=True)

    _logger.debug("Stopping loop")
    loop.stop()


def _exit_handler(close_loop):
    _logger.debug("Received stop signal")
    asyncio.ensure_future(close_loop())


async def _start_monitor(monitor, close_loop):
    if not monitor:
        return

    try:
        await monitor.start()
    except:
        _logger.warning("Error starting monitor", exc_info=True)
        await close_loop()


def run_mqtt_broker(conf, disable_monitor):
    assert sh.Command("mosquitto")

    loop = asyncio.get_event_loop()
    loop.set_exception_handler(_loop_ex_handler)

    monitor = NodeMonitor(
        redis_url=conf.redis_url) if not disable_monitor else None

    close_loop = functools.partial(
        _close_loop,
        loop=loop,
        monitor=monitor)

    exit_handler = functools.partial(
        _exit_handler,
        close_loop=close_loop)

    start_monitor = functools.partial(
        _start_monitor,
        monitor=monitor,
        close_loop=close_loop)

    for name in {"SIGINT", "SIGTERM"}:
        loop.add_signal_handler(getattr(signal, name), exit_handler)

    proc = None

    try:
        asyncio.ensure_future(start_monitor())
        proc = _run_broker(conf=conf, loop=loop, close_loop=close_loop)
        _logger.debug("Running loop")
        loop.run_forever()
    finally:
        _logger.debug("Closing loop")
        loop.close()
        _terminate_broker(proc=proc)



import enum


class Labels(enum.Enum):
    WOTEMU_NETWORK = "org.fundacionctic.wotemu.net"
    WOTEMU_GATEWAY = "org.fundacionctic.wotemu.gw"
    WOTEMU_NODE = "org.fundacionctic.wotemu.node"
    WOTEMU_BROKER = "org.fundacionctic.wotemu.broker"
    WOTEMU_REDIS = "org.fundacionctic.wotemu.redis"
    WOTEMU_SERVICE = "org.fundacionctic.wotemu.service"
    WOTEMU_SERVICE_NETWORK = "org.fundacionctic.wotemu.servicenet"


class RedisPrefixes(enum.Enum):
    NAMESPACE = "wotemu"
    PACKET = "packet"
    THING = "thing"
    SYSTEM = "system"
    INFO = "info"
    BENCHMARK = "benchmark"
    SNAPSHOT = "snapshot"
    COMPOSE = "compose"


class NetworkConditions(enum.Enum):
    GPRS = "GPRS"
    EDGE = "EDGE"
    REGULAR_3G = "REGULAR_3G"
    FAST_3G = "FAST_3G"
    LTE = "LTE"
    WIFI = "WIFI"
    CABLE = "CABLE"


class NodePlatforms(enum.Enum):
    CONSTRAINED = "CONSTRAINED"
    GATEWAY = "GATEWAY"
    CLOUD = "CLOUD"
    UNCONSTRAINED = "UNCONSTRAINED"


_DELAY = "delay --time {latency} --jitter {jitter} --distribution normal"
_RATE = "rate --rate {rate}"

NETEM_CONDITIONS = {
    NetworkConditions.GPRS: [
        _DELAY.format(latency=700, jitter=100),
        _RATE.format(rate="50kbit")
    ],
    NetworkConditions.EDGE: [
        _DELAY.format(latency=700, jitter=100),
        _RATE.format(rate="100kbit")
    ],
    NetworkConditions.REGULAR_3G: [
        _DELAY.format(latency=300, jitter=150),
        _RATE.format(rate="1500kbit")
    ],
    NetworkConditions.FAST_3G: [
        _DELAY.format(latency=150, jitter=50),
        _RATE.format(rate="4000kbit")
    ],
    NetworkConditions.LTE: [
        _DELAY.format(latency=40, jitter=10),
        _RATE.format(rate="15mbit")
    ],
    NetworkConditions.WIFI: [
        _DELAY.format(latency=25, jitter=5),
        _RATE.format(rate="50mbit")
    ],
    NetworkConditions.CABLE: [
        _DELAY.format(latency=5, jitter=5),
        _RATE.format(rate="100mbit")
    ]
}

assert set(NETEM_CONDITIONS.keys()) == set(NetworkConditions)


class BuiltinApps(enum.Enum):
    CLOCK = "wotemu_clock"
    ERROR = "wotemu_error"
    READER = "wotemu_reader"
    SUBSCRIBER = "wotemu_subscriber"
    WORKER = "wotemu_worker"
    CALLER = "wotemu_caller"
    MONGO_HISTORIAN = "wotemu_mongo_historian"
    CAMERA = "wotemu_camera"
    DETECTOR = "wotemu_detector"


BUILTIN_APPS_MODULES = {
    BuiltinApps.CLOCK: "wotemu.apps.clock",
    BuiltinApps.ERROR: "wotemu.apps.error",
    BuiltinApps.READER: "wotemu.apps.reader",
    BuiltinApps.SUBSCRIBER: "wotemu.apps.subscriber",
    BuiltinApps.WORKER: "wotemu.apps.worker",
    BuiltinApps.CALLER: "wotemu.apps.caller",
    BuiltinApps.MONGO_HISTORIAN: "wotemu.apps.historian",
    BuiltinApps.CAMERA: "wotemu.apps.camera",
    BuiltinApps.DETECTOR: "wotemu.apps.detector"
}

assert set(BUILTIN_APPS_MODULES.keys()) == set(BuiltinApps)



import importlib
import logging
import os
import sys

from wotemu.utils import import_func

_logger = logging.getLogger(__name__)


def _default_output_path(path):
    path_root, path_base = os.path.split(path)
    name, _ext = os.path.splitext(path_base)
    return os.path.join(path_root, "{}.yml".format(name))


def build_compose(conf, path, output, func):
    topology_func = import_func(path, func)
    _logger.info("Imported topology function: %s", topology_func)
    output = output if output else _default_output_path(path)
    _logger.info("Writing Compose file to: %s", output)
    compose_yaml = topology_func().to_compose_yaml()

    with open(output, "w") as fh:
        fh.write(compose_yaml)



import asyncio
import json
import random

import pytest
from tests.conftest import TD_EXAMPLE
from wotpy.protocols.enums import InteractionVerbs
from wotpy.wot.td import ThingDescription

from wotemu.wotpy.things import ConsumedThing, ExposedThing
from wotemu.wotpy.wot import wot_entrypoint

MAX_DUMMY_SLEEP_SECS = 0.15
WOT_HOSTNAME = "127.0.0.1"


async def dummy_sleep():
    await asyncio.sleep(random.uniform(0, MAX_DUMMY_SLEEP_SECS))


async def max_sleep():
    await asyncio.sleep(MAX_DUMMY_SLEEP_SECS * 2)


def decorated_wotpy_params():
    params = [
        "port_http",
        "port_ws"
    ]

    try:
        from wotpy.protocols.coap.server import CoAPServer
        params.append("port_coap")
    except NotImplementedError:
        pass

    return params


@pytest.fixture(params=decorated_wotpy_params())
def decorated_wotpy(request, unused_tcp_port_factory):
    exposed_data = []

    async def exposed_cb(data):
        await dummy_sleep()
        exposed_data.append(data)

    consumed_data = []

    async def consumed_cb(data):
        await dummy_sleep()
        consumed_data.append(data)

    port_catalogue = unused_tcp_port_factory()

    wot_kwargs = {
        request.param: unused_tcp_port_factory()
    }

    wot = wot_entrypoint(
        port_catalogue=port_catalogue,
        hostname=WOT_HOSTNAME,
        exposed_cb=exposed_cb,
        consumed_cb=consumed_cb,
        **wot_kwargs)

    return {
        "wot": wot,
        "exposed_data": exposed_data,
        "consumed_data": consumed_data
    }


async def build_things(wot):
    exposed_thing = wot.produce(model=json.dumps(TD_EXAMPLE))
    exposed_thing.expose()
    await exposed_thing.servient.start()

    td_str = ThingDescription.from_thing(exposed_thing.thing).to_str()
    consumed_thing = wot.consume(td_str)

    return exposed_thing, consumed_thing


@pytest.mark.asyncio
async def text_exposed_decorated(event_loop, decorated_wotpy):
    wot = decorated_wotpy["wot"]
    exposed_thing = wot.produce(model=json.dumps(TD_EXAMPLE))
    assert isinstance(exposed_thing, ExposedThing)


@pytest.mark.asyncio
async def test_consumed_decorated(event_loop, decorated_wotpy):
    wot = decorated_wotpy["wot"]
    consumed_thing = wot.consume(json.dumps(TD_EXAMPLE))
    assert isinstance(consumed_thing, ConsumedThing)


@pytest.mark.asyncio
async def test_exposed_rw_property(decorated_wotpy):
    wot = decorated_wotpy["wot"]
    exposed_data = decorated_wotpy["exposed_data"]

    _, consumed_thing = await build_things(wot)

    assert len(exposed_data) == 0

    val = int(random.random() * 100)
    await consumed_thing.properties["testProp"].write(val)
    read_val = await consumed_thing.properties["testProp"].read()
    assert read_val == val

    await max_sleep()

    assert any(
        item.get("verb") == InteractionVerbs.READ_PROPERTY and
        item.get("result")
        for item in exposed_data)

    assert any(
        item.get("verb") == InteractionVerbs.WRITE_PROPERTY
        for item in exposed_data)


@pytest.mark.asyncio
async def test_exposed_error(decorated_wotpy):
    wot = decorated_wotpy["wot"]
    exposed_data = decorated_wotpy["exposed_data"]

    async def read_handler():
        raise Exception

    exposed_thing, _ = await build_things(wot)
    exposed_thing.set_property_read_handler("testProp", read_handler)

    with pytest.raises(Exception):
        await exposed_thing.properties["testProp"].read()

    await max_sleep()

    assert any(
        item.get("verb") == InteractionVerbs.READ_PROPERTY and
        item.get("error") and
        item.get("result", None) is None
        for item in exposed_data)


@pytest.mark.asyncio
async def test_exposed_observe_property(event_loop, decorated_wotpy):
    wot = decorated_wotpy["wot"]
    consumed_data = decorated_wotpy["consumed_data"]

    exposed_thing, consumed_thing = await build_things(wot)

    assert len(consumed_data) == 0

    val = int(random.random() * 100)

    async def write_loop():
        try:
            while True:
                exposed_thing.properties["testProp"].write(val)
                await asyncio.sleep(0.1)
        except asyncio.CancelledError:
            pass

    sub_future = event_loop.create_future()

    def on_next(item):
        if not sub_future.done():
            sub_future.set_result(item.data.value)

    sub = consumed_thing.properties["testProp"].subscribe(on_next=on_next)
    write_task = event_loop.create_task(write_loop())
    sub_value = await sub_future
    assert sub_value == val
    write_task.cancel()

    await max_sleep()

    assert any(
        item.get("verb") == InteractionVerbs.OBSERVE_PROPERTY and
        item.get("event") == "on_next" and
        item.get("item")
        for item in consumed_data)

    sub.dispose()
    await exposed_thing.servient.shutdown()
    await max_sleep()



import asyncio
import json
import logging
import os
import time

import aioredis
import docker
import numpy
from wotemu.enums import RedisPrefixes
from wotemu.topology.compose import ENV_KEY_CPU_SPEED, ENV_KEY_NODE_ID
from wotemu.topology.cpu import get_cpu_core_scale, get_cpu_core_speed_poly
from wotemu.utils import get_current_container_id, ping_docker

_CPU_PERIOD = 100000
_KEY_CPU_PREFIX = "cpu"
_EXPIRE_SECS = 3600

_logger = logging.getLogger(__name__)


def _get_cpu_poly_key():
    node_id = os.environ.get(ENV_KEY_NODE_ID, None)

    if node_id is None:
        raise RuntimeError(f"Undefined ${ENV_KEY_NODE_ID}")

    return "{}:{}:{}".format(
        RedisPrefixes.NAMESPACE.value,
        RedisPrefixes.BENCHMARK.value,
        node_id)


def _local_cpu_poly():
    return get_cpu_core_speed_poly(cache=True)


async def _get_redis_poly(redis, key):
    _logger.debug("GET %s", key)
    val = await redis.get(key)

    if not val:
        return None

    coeffs = json.loads(val)

    return numpy.poly1d(coeffs) if coeffs else None


async def _set_redis_poly(redis, key, poly):
    val = json.dumps(poly.coeffs.tolist())
    _logger.debug("SET %s %s", key, val)
    await redis.set(key, val, expire=_EXPIRE_SECS)


async def _should_build_poly(redis, key):
    return not (await redis.get(key))


async def _lock_poly(redis, key):
    val = json.dumps(None)
    _logger.debug("SET %s %s", key, val)
    await redis.set(key, val, expire=_EXPIRE_SECS, exist=redis.SET_IF_NOT_EXIST)


async def _get_cpu_poly(redis_url, timeout=300, wait=2):
    key = _get_cpu_poly_key()

    if not key:
        raise ValueError("Undefined CPU speed poly Redis key")

    _logger.debug("Connecting to Redis: %s", redis_url)
    redis = await aioredis.create_redis_pool(redis_url)

    if (await _should_build_poly(redis, key)):
        _logger.info("Building CPU speed poly locally")
        await _lock_poly(redis, key)
        poly = _local_cpu_poly()
        await _set_redis_poly(redis, key, poly)
        return poly

    _logger.info("Waiting for CPU speed poly on Redis")

    ini = time.time()

    def _raise_timeout():
        if timeout is None:
            return

        if (time.time() - ini) >= timeout:
            raise Exception("Timeout waiting for Redis poly")

    while True:
        poly = await _get_redis_poly(redis, key)

        if poly:
            return poly
        else:
            await asyncio.sleep(wait)

        _raise_timeout()


def update_limits(conf, docker_url, speed, timeout, wait, local):
    ping_docker(docker_url=docker_url)

    cpu_poly = None

    if not local:
        _logger.debug((
            "Attempting to build the CPU speed poly "
            "in a distributed fashion using Redis as a "
            "central cache to avoid overloading the host "
            "with multiple parallel CPU speed benchmarks"
        ))

        try:
            loop = asyncio.get_event_loop()
            cpu_poly_fut = _get_cpu_poly(conf.redis_url, timeout, wait)
            cpu_poly = loop.run_until_complete(cpu_poly_fut)
        except Exception as ex:
            _logger.warning("Error getting CPU poly from Redis: %s", repr(ex))

    if not cpu_poly:
        _logger.info("Forced to build local CPU speed poly")
        cpu_poly = _local_cpu_poly()

    _logger.debug("Calculating the CPU scale to match target speed: %s", speed)
    speed_scale = get_cpu_core_scale(speed, core_poly=cpu_poly)

    _logger.debug("Getting current container to update CPU limits")
    cid = get_current_container_id()
    docker_client = docker.DockerClient(base_url=docker_url)
    curr_container = docker_client.containers.get(cid)

    update_kwargs = {
        "cpu_period": _CPU_PERIOD,
        "cpu_quota": int(_CPU_PERIOD * speed_scale)
    }

    _logger.info("Container update parameters: %s", update_kwargs)
    result = curr_container.update(**update_kwargs)

    _logger.log(
        logging.WARNING if result.get("Warnings", None) else logging.DEBUG,
        "Container update result: %s",
        result)



import functools
import json
import logging
import re
import socket
from datetime import datetime, timezone

import aioredis
import netaddr
import numpy as np
import pandas as pd
from wotemu.enums import RedisPrefixes
from wotemu.topology.compose import ENV_KEY_SERVICE_NAME

_IFACE_LO = "lo"
_DOCKER_TIME_REGEX = r"^(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}).(\d+)Z$"
_DOCKER_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
_STATE_RUNNING = "running"

_logger = logging.getLogger(__name__)


def _parse_docker_time(val):
    match = re.match(_DOCKER_TIME_REGEX, val)

    if not match:
        return None

    dtime = datetime.strptime(match.group(1), _DOCKER_TIME_FORMAT)
    dtime = dtime.replace(tzinfo=timezone.utc)
    dtime = dtime.replace(microsecond=int(match.group(2)[:6]))

    return dtime


def _is_task_running(task_dict):
    return task_dict["DesiredState"] == _STATE_RUNNING


def _is_task_error(task_dict):
    is_running = _is_task_running(task_dict)

    exit_code = task_dict.\
        get("Status", {}).\
        get("ContainerStatus", {}).\
        get("ExitCode", 0)

    return not is_running and exit_code != 0


def _explode_mapper(val, key):
    if val is np.nan:
        return np.nan

    try:
        return val[key]
    except:
        return np.nan


def _unique_keys_dict_column(df, col):
    dict_filter = df[col].apply(lambda x: isinstance(x, dict))
    keys = df[dict_filter][col].apply(tuple).unique()

    # Flatten list of tuples:
    # https://stackoverflow.com/a/10636583
    keys = set(sum(keys, ()))

    return keys


def explode_dict_column(df, col):
    if col not in df:
        return

    keys = _unique_keys_dict_column(df, col)

    for key in keys:
        mapper = functools.partial(_explode_mapper, key=key)
        df[f"{col}_{key}"] = df[col].apply(mapper)


class ReportDataRedisReader:
    def __init__(self, redis_url):
        self._redis_url = redis_url
        self._client = None

    async def connect(self):
        await self.close()
        self._client = await aioredis.create_redis_pool(self._redis_url)

    async def close(self):
        if self._client is None:
            return

        try:
            self._client.close()
            await self._client.wait_closed()
        except Exception as ex:
            _logger.warning("Error closing connection: %s", ex)
        finally:
            self._client = None

    async def _get_zrange_df(self, key):
        members = await self._client.zrange(key=key)
        rows = [json.loads(item) for item in members]

        for row in rows:
            row_date = datetime.fromtimestamp(row["time"], timezone.utc)
            row.update({"date": row_date})

        df = pd.DataFrame(rows)

        if "date" in df:
            df.set_index("date", inplace=True)

        return df

    async def get_tasks(self):
        pattern = "{}:{}:*".format(
            RedisPrefixes.NAMESPACE.value,
            RedisPrefixes.INFO.value)

        keys = await self._client.keys(pattern=pattern)

        return {key.decode().split(":")[-1] for key in keys}

    async def get_info(self, task, latest=False):
        key = "{}:{}:{}".format(
            RedisPrefixes.NAMESPACE.value,
            RedisPrefixes.INFO.value,
            task)

        members = await self._client.zrange(key=key)
        rows = [json.loads(item) for item in members]
        rows.sort(key=lambda row: row["time"])

        if latest:
            return rows[-1] if len(rows) > 0 else None
        else:
            return rows

    async def get_info_map(self):
        task_keys = await self.get_tasks()

        ret = {
            task_key: await self.get_info(task_key, latest=True)
            for task_key in task_keys
        }

        return {task_key: info for task_key, info in ret.items() if info}

    async def get_compose_dict(self):
        key = "{}:{}".format(
            RedisPrefixes.NAMESPACE.value,
            RedisPrefixes.COMPOSE.value)

        members = await self._client.zrange(key=key, start=-1, stop=-1)

        if not members or not len(members):
            return None

        return json.loads(members[-1])

    async def get_system_df(self, task):
        key = "{}:{}:{}".format(
            RedisPrefixes.NAMESPACE.value,
            RedisPrefixes.SYSTEM.value,
            task)

        return await self._get_zrange_df(key=key)

    async def get_packet_df(self, task, extended=False):
        pattern = "{}:{}:*:{}".format(
            RedisPrefixes.NAMESPACE.value,
            RedisPrefixes.PACKET.value,
            task)

        packet_keys = await self._client.keys(pattern=pattern)

        if len(packet_keys) == 0:
            _logger.debug(
                "No packet keys found for task '%s': Undefined packet DF", task)

            return None

        dfs = []

        for key in packet_keys:
            iface = key.decode().split(":")[2]
            df_iface = await self._get_zrange_df(key=key)
            df_iface["iface"] = iface
            df_iface.set_index(["iface"], append=True, inplace=True)
            dfs.append(df_iface)

        df = pd.concat(dfs)
        df.sort_index(inplace=True)

        if extended:
            df = await self.extend_packet_df(df_packet=df)

        return df

    async def extend_packet_df(self, df_packet, df_address=None, df_vip=None):
        df_address = df_address if df_address else await self.get_address_df()
        df_address = df_address.reset_index().drop(columns=["iface", "date"])

        df_vip = df_vip if df_vip else await self.get_service_vip_df()
        df_vip = df_vip.reset_index().drop(columns=["date", "task"])

        df_packet = df_packet.reset_index()

        df = pd.merge(
            df_packet, df_address,
            how="left", left_on="src", right_on="address")

        df = df.drop(columns=["address"]).rename(columns={
            "task": "src_task",
            "service": "src_service",
            "network": "src_network"
        })

        df = pd.merge(
            df, df_address,
            how="left", left_on="dst", right_on="address")

        df = df.drop(columns=["address"]).rename(columns={
            "task": "dst_task",
            "service": "dst_service",
            "network": "dst_network"
        })

        df = pd.merge(
            df, df_vip,
            how="left", left_on="src", right_on="vip")

        df = df.drop(columns=["vip", "network"]).rename(columns={
            "service": "src_vip_service"
        })

        df = pd.merge(
            df, df_vip,
            how="left", left_on="dst", right_on="vip")

        df = df.drop(columns=["vip", "network"]).rename(columns={
            "service": "dst_vip_service"
        })

        df["src_service"] = df["src_service"].fillna(
            value=df["src_vip_service"])

        df["dst_service"] = df["dst_service"].fillna(
            value=df["dst_vip_service"])

        df["network"] = df["src_network"].combine_first(df["dst_network"])

        df = df.drop(columns=[
            "src_vip_service",
            "dst_vip_service",
            "src_network",
            "dst_network"
        ])

        df.set_index(["date", "iface"], inplace=True)

        nan_series = df.isna().sum() / len(df)
        nan_series = nan_series[nan_series > 0]

        if len(nan_series) > 0:
            warn_cols = {"network", "len", "src", "dst", "proto", "transport"}
            is_warn = len(warn_cols.intersection(set(nan_series.index))) > 0

            _logger.log(
                logging.WARNING if is_warn else logging.DEBUG,
                "NaN ratio for extended packet DF (len: %s):\n%s",
                len(df), nan_series)

        return df

    async def get_thing_df(self, task):
        key = "{}:{}:{}".format(
            RedisPrefixes.NAMESPACE.value,
            RedisPrefixes.THING.value,
            task)

        df = await self._get_zrange_df(key=key)

        for col in ["thing", "name", "verb"]:
            if col in df:
                df.set_index([col], append=True, inplace=True)

        return df

    def _find_network(self, info_item, address):
        ip_addr = netaddr.IPAddress(address)
        networks_cidr = info_item.get("networks_cidr", {})

        try:
            return next(
                net_name
                for net_name, cidr_list in networks_cidr.items()
                for cidr in cidr_list
                if ip_addr in netaddr.IPNetwork(cidr))
        except StopIteration:
            return np.nan

    async def get_address_df(self, tasks=None):
        tasks = tasks if tasks else await self.get_tasks()

        rows = [
            {
                "date": datetime.fromtimestamp(info_item["time"], timezone.utc),
                "task": task,
                "service": info_item.get("env", {}).get(ENV_KEY_SERVICE_NAME),
                "iface": iface,
                "address": iface_item["address"],
                "network": self._find_network(info_item, iface_item["address"])
            }
            for task in tasks
            for info_item in await self.get_info(task=task)
            for iface, iface_items in info_item.get("net", {}).items() if iface != _IFACE_LO
            for iface_item in iface_items if iface_item.get("family") == socket.AF_INET
        ]

        df = pd.DataFrame(rows)
        df.set_index(["date", "task", "iface"], inplace=True)

        return df

    async def get_service_vip_df(self, tasks=None):
        tasks = tasks if tasks else await self.get_tasks()

        rows = [
            {
                "date": datetime.fromtimestamp(info_item["time"], timezone.utc),
                "task": task,
                "service": info_item.get("env", {}).get(ENV_KEY_SERVICE_NAME),
                "network": net_name,
                "vip": vip
            }
            for task in tasks
            for info_item in await self.get_info(task=task)
            for net_name, vip in info_item.get("service_vips", {}).items()
        ]

        df = pd.DataFrame(rows)
        df.set_index(["date", "task", "network"], inplace=True)

        return df

    async def get_service_traffic_df(self, tasks=None, inbound=True):
        tasks = tasks if tasks else await self.get_tasks()

        col_task = "src_task" if inbound else "dst_task"
        col_service = "dst_service" if inbound else "src_service"

        dfs = {
            task: await self.get_packet_df(task, extended=True)
            for task in tasks
        }

        dfs = {
            task: df[(df[col_task] == task) & df[col_service].notna()]
            for task, df in dfs.items()
            if df is not None
        }

        dfs = {
            task: df.groupby(col_service)["len"].sum().to_frame()
            for task, df in dfs.items()
        }

        for task, df in dfs.items():
            df.reset_index(inplace=True)
            df[col_task] = task

        return pd.concat(dfs.values(), ignore_index=True) if len(dfs) else None

    async def _get_task_key_map(self):
        task_keys = await self.get_tasks()

        task_infos = {
            task_key: await self.get_info(task_key)
            for task_key in task_keys
        }

        task_infos = {
            task_key: info_items[-1]
            for task_key, info_items in task_infos.items()
            if len(info_items) > 0
        }

        return {
            info_item["task_id"]: task_key
            for task_key, info_item in task_infos.items()
            if info_item.get("container_id")
        }

    async def get_snapshot_df(self):
        key = "{}:{}".format(
            RedisPrefixes.NAMESPACE.value,
            RedisPrefixes.SNAPSHOT.value)

        members = await self._client.zrange(key=key, start=-1, stop=-1, withscores=True)

        if len(members) == 0:
            _logger.warning("Could not find snapshot data")
            return None

        snapshot = json.loads(members[-1][0])
        snapshot_dtime = datetime.fromtimestamp(members[-1][1], timezone.utc)

        task_key_map = await self._get_task_key_map()

        rows = [
            {
                "desired_state": item["task"].get("DesiredState"),
                "is_running": _is_task_running(item["task"]),
                "is_error": _is_task_error(item["task"]),
                "task": task_key_map.get(task_id),
                "task_id": task_id,
                "node_id": item["task"].get("NodeID"),
                "service_id": item["task"].get("ServiceID"),
                "created_at": _parse_docker_time(item["task"].get("CreatedAt")),
                "updated_at": _parse_docker_time(item["task"].get("UpdatedAt")),
                "container_id": item["task"].get("Status", {}).get("ContainerStatus", {}).get("ContainerID"),
                "logs": item["logs"],
                "stopped_at": snapshot_dtime
            }
            for task_id, item in snapshot.items()
        ]

        df = pd.DataFrame(rows)

        return df



import json
import logging
import os

import aioredis
import wotemu.config
from wotemu.enums import RedisPrefixes

_logger = logging.getLogger(__name__)


async def redis_from_env():
    env_config = wotemu.config.get_env_config()

    if not env_config.redis_url:
        raise Exception("Undefined Redis URL")

    return await aioredis.create_redis_pool(env_config.redis_url)


async def redis_thing_callback(data, client=None):
    try:
        redis = None
        redis = client if client else await redis_from_env()

        key = "{}:{}:{}".format(
            RedisPrefixes.NAMESPACE.value,
            RedisPrefixes.THING.value,
            data["host"])

        score = data["time"]
        member = json.dumps(data)
        await redis.zadd(key=key, score=score, member=member)
    except Exception as ex:
        _logger.warning("Error in Redis callback: %s", ex)
    finally:
        if client is None and redis:
            redis.close()
            await redis.wait_closed()



import collections
import enum
import logging
import os
import pprint

DEFAULT_DOCKER_SOCKET = "/var/run/docker.sock"
DEFAULT_HOST_REDIS = "redis"
DEFAULT_HOST_DOCKER_PROXY = "docker_api_proxy"

_DEFAULT_PORT_CATALOGUE = 9090
_DEFAULT_PORT_HTTP = 80
_DEFAULT_PORT_WS = 81
_DEFAULT_PORT_COAP = 5683
_DEFAULT_PORT_MQTT = 1883
_DEFAULT_REDIS_URL = "redis://{}".format(DEFAULT_HOST_REDIS)
_DEFAULT_DOCKER_PROXY_URL = "tcp://{}:2375/".format(DEFAULT_HOST_DOCKER_PROXY)

_logger = logging.getLogger(__name__)


EnvConfig = collections.namedtuple(
    "EnvConfig",
    [
        "port_catalogue",
        "port_http",
        "port_coap",
        "port_ws",
        "port_mqtt",
        "mqtt_broker_host",
        "mqtt_url",
        "redis_url",
        "docker_proxy_url",
        "other_ports_tcp",
        "other_ports_udp"
    ])


class ConfigVars(enum.Enum):
    PORT_CATALOGUE = "PORT_CATALOGUE"
    PORT_HTTP = "PORT_HTTP"
    PORT_WS = "PORT_WS"
    PORT_COAP = "PORT_COAP"
    PORT_MQTT = "PORT_MQTT"
    MQTT_BROKER_HOST = "MQTT_BROKER_HOST"
    REDIS_URL = "REDIS_URL"
    DOCKER_PROXY_URL = "DOCKER_PROXY_URL"
    OTHER_PORTS_TCP = "OTHER_PORTS_TCP"
    OTHER_PORTS_UDP = "OTHER_PORTS_UDP"


DEFAULT_CONFIG_VARS = {
    ConfigVars.PORT_CATALOGUE: _DEFAULT_PORT_CATALOGUE,
    ConfigVars.PORT_HTTP: _DEFAULT_PORT_HTTP,
    ConfigVars.PORT_WS: _DEFAULT_PORT_WS,
    ConfigVars.PORT_COAP: _DEFAULT_PORT_COAP,
    ConfigVars.PORT_MQTT: _DEFAULT_PORT_MQTT,
    ConfigVars.MQTT_BROKER_HOST: None,
    ConfigVars.REDIS_URL: _DEFAULT_REDIS_URL,
    ConfigVars.DOCKER_PROXY_URL: _DEFAULT_DOCKER_PROXY_URL,
    ConfigVars.OTHER_PORTS_TCP: None,
    ConfigVars.OTHER_PORTS_UDP: None
}


def _getenv_int(name, default):
    try:
        return int(os.getenv(name, default))
    except:
        _logger.warning(
            "Unexpected int value (%s) in variable %s: Using default (%s)",
            os.getenv(name), name, default)

        return default


def _parse_ports(val):
    try:
        return [int(item) for item in val.split(",")]
    except:
        _logger.warning(f"Error parsing ports: {val}", exc_info=True)
        return None


def get_env_config():
    port_catalogue = _getenv_int(
        ConfigVars.PORT_CATALOGUE.value,
        DEFAULT_CONFIG_VARS.get(ConfigVars.PORT_CATALOGUE))

    port_http = _getenv_int(
        ConfigVars.PORT_HTTP.value,
        DEFAULT_CONFIG_VARS.get(ConfigVars.PORT_HTTP))

    port_ws = _getenv_int(
        ConfigVars.PORT_WS.value,
        DEFAULT_CONFIG_VARS.get(ConfigVars.PORT_WS))

    port_coap = _getenv_int(
        ConfigVars.PORT_COAP.value,
        DEFAULT_CONFIG_VARS.get(ConfigVars.PORT_COAP))

    port_mqtt = _getenv_int(
        ConfigVars.PORT_MQTT.value,
        DEFAULT_CONFIG_VARS.get(ConfigVars.PORT_MQTT))

    mqtt_broker_host = os.getenv(
        ConfigVars.MQTT_BROKER_HOST.value,
        DEFAULT_CONFIG_VARS.get(ConfigVars.MQTT_BROKER_HOST))

    redis_url = os.getenv(
        ConfigVars.REDIS_URL.value,
        DEFAULT_CONFIG_VARS.get(ConfigVars.REDIS_URL))

    docker_proxy_url = os.getenv(
        ConfigVars.DOCKER_PROXY_URL.value,
        DEFAULT_CONFIG_VARS.get(ConfigVars.DOCKER_PROXY_URL))

    other_ports_tcp = os.getenv(
        ConfigVars.OTHER_PORTS_TCP.value,
        DEFAULT_CONFIG_VARS.get(ConfigVars.OTHER_PORTS_TCP))

    other_ports_udp = os.getenv(
        ConfigVars.OTHER_PORTS_UDP.value,
        DEFAULT_CONFIG_VARS.get(ConfigVars.OTHER_PORTS_UDP))

    other_ports_tcp = other_ports_tcp and _parse_ports(other_ports_tcp)
    other_ports_udp = other_ports_udp and _parse_ports(other_ports_udp)

    mqtt_url = None

    if port_mqtt and mqtt_broker_host:
        mqtt_url = "mqtt://{}:{}".format(mqtt_broker_host, port_mqtt)

    config = EnvConfig(
        port_catalogue=port_catalogue,
        port_http=port_http,
        port_ws=port_ws,
        port_coap=port_coap,
        port_mqtt=port_mqtt,
        mqtt_broker_host=mqtt_broker_host,
        mqtt_url=mqtt_url,
        redis_url=redis_url,
        docker_proxy_url=docker_proxy_url,
        other_ports_tcp=other_ports_tcp,
        other_ports_udp=other_ports_udp)

    return config


def log_config():
    conf_env = {key.value: os.getenv(key.value, None) for key in ConfigVars}
    _logger.debug("Configuration environment:\n%s", pprint.pformat(conf_env))
    _logger.debug("Current configuration: %s", get_env_config())



from os import path

import setuptools

from wotemu.__version__ import __version__


def _get_readme():
    curr_dir = path.abspath(path.dirname(__file__))

    with open(path.join(curr_dir, "README.md")) as fh:
        return fh.read()


setuptools.setup(
    name="wotemu",
    version=__version__,
    keywords='wot iot edge emulator',
    author='Andres Garcia Mangas',
    author_email='andres.garcia@fundacionctic.org',
    description="An emulator for Python applications to help in the design of IoT deployments based on the edge computing model",
    long_description=_get_readme(),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX :: Linux"
    ],
    entry_points={
        "console_scripts": [
            "wotemu=wotemu.cli.main:cli"
        ]
    },
    python_requires='>=3.6',
    install_requires=[
        "docker>=4.1.0,<5.0",
        "coloredlogs>=14.0,<15.0",
        "netaddr>=0.7.19,<0.8.0",
        "netifaces>=0.10.9,<0.11.0",
        "Click>=7.0,<8.0",
        "sh>=1.12.14,<2.0",
        "wotpy==0.14.5",
        "aioredis>=1.3,<2.0",
        "tornado>=5.1,<6.0",
        "PyYAML>=5.3,<6.0",
        "inflection>=0.4.0,<0.5.0",
        "numpy>=1.15.0,<2.0",
        "deepmerge>=0.1.0,<0.2.0",
        "pyshark>=0.4.2,<0.5.0",
        "psutil>=5.6.0,<6.0",
        "plotly>=4.11,<5.0",
        "pandas>=1.1,<2.0",
        "lxml>=4.5,<5.0",
        "websockets>=8.0,<9.0"
    ],
    extras_require={
        "dev": [
            "autopep8>=1.5,<2.0",
            "pylint>=2.0,<3.0",
            "rope>=0.16.0,<1.0",
            "pytest>=5.0,<6.0",
            "pytest-asyncio>=0.10.0,<1.0",
            "docker>=4.0,<5.0",
            "html5lib>=1.1,<2.0",
            "bumpversion>=0.5.3,<0.6.0"
        ],
        "apps": [
            "motor>=2.3,<3.0",
            "opencv-python>=4.5,<4.6",
            "face-recognition>=1.3,<1.4"
        ]
    }
)



"""
Invokes all the actions from the Thing 
passed as argument in a periodic fashion.
"""

import asyncio
import json
import logging
import pprint
import random

from wotemu.utils import consume_from_catalogue, wait_node

_logger = logging.getLogger(__name__)


def _log_result(res, max_len=1024):
    res_str = pprint.pformat(res)

    if len(res_str) > max_len:
        res_str = f"{res_str[:max_len]} [...]"

    _logger.debug("Invocation result:\n%s", res_str)


async def _invoke(consumed_thing, name, params, timeout, max_log_len=512):
    if name in params and not params[name]:
        _logger.debug("Skipping invocation for action: %s", name)
        return

    try:
        params = params.get(name, None)
        _logger.debug("Invoking %s with params: %s", name, params)
        res = await consumed_thing.actions[name].invoke(params, timeout=timeout)
        _log_result(res)
    except Exception:
        _logger.warning("Invocation error (%s)", name, exc_info=True)


async def app(wot, conf, loop, servient_host, thing_id, params=None, timeout=300, lambd=2):
    try:
        params = json.loads(params)
    except:
        _logger.debug("Could not parse JSON-serialized params")

    params = params or {}
    lambd = float(lambd)

    _logger.info(
        "Starting caller app for remote %s :: %s (lambda=%s)",
        servient_host, thing_id, lambd)

    await wait_node(conf=conf, name=servient_host)

    consumed_thing = await consume_from_catalogue(
        wot=wot,
        port_catalogue=conf.port_catalogue,
        servient_host=servient_host,
        thing_id=thing_id)

    while True:
        for name in consumed_thing.actions:
            await _invoke(
                consumed_thing=consumed_thing,
                name=name,
                params=params,
                timeout=timeout)

        await asyncio.sleep(random.expovariate(lambd))



import logging
import pprint
import re
import warnings

import docker
import numpy

_SYSBENCH_IMAGE = "severalnines/sysbench"
_SYSBENCH_THREADS = 1
_SYSBENCH_TIME = 10
_CPU_PERIOD = 100000
_REGEX_SPEED = r"events\sper\ssecond:\s*(\d+\.\d+)"

_cache = dict()
_logger = logging.getLogger(__name__)


def get_cpu_core_speed(cpus):
    client = docker.from_env()

    command = "sysbench --threads={} --time={} cpu run".format(
        _SYSBENCH_THREADS,
        _SYSBENCH_TIME)

    cpu_quota = int(_CPU_PERIOD * cpus)

    _logger.info("Running for [cpus=%s]: %s", cpus, command)

    output = client.containers.run(
        remove=True,
        image=_SYSBENCH_IMAGE,
        command=command,
        cpu_period=_CPU_PERIOD,
        cpu_quota=cpu_quota)

    output = output.decode()

    _logger.debug("[cpus=%s] %s\n%s", cpus, command, output)

    speed_match = re.search(_REGEX_SPEED, output)

    if len(speed_match.groups()) != 1:
        raise Exception("CPU speed not found in sysbench output")

    cpu_speed = float(speed_match.group(1))

    _logger.debug("CPU speed: %s", cpu_speed)

    return cpu_speed


def get_cpu_core_speed_poly(num=6, start=0.1, stop=1.0, cache=True):
    cache_key = (num, start, stop)

    if start <= 0:
        raise ValueError("Parameter 'start' should be > 0")

    if cache and _cache.get(cache_key):
        _logger.debug("Using cached CPU core speed poly")
        return _cache.get(cache_key)

    _logger.info("Building CPU core speed poly")

    cpu_samples = numpy.linspace(start=start, stop=stop, num=num)

    speeds = [(0.0, 0.0)] + [
        (cpus, get_cpu_core_speed(cpus))
        for cpus in cpu_samples
    ]

    _logger.debug("CPU speed dataset:\n%s", pprint.pformat(speeds))

    cpu_ratio = [item[0] for item in speeds]
    cpu_speed = [item[1] for item in speeds]
    coeffs = numpy.polyfit(cpu_ratio, cpu_speed, deg=1)
    poly = numpy.poly1d(coeffs)

    _logger.debug("CPU speed poly: %s", poly)

    _cache[cache_key] = poly

    return poly


def get_cpu_core_scale(target_core_speed, core_poly=None):
    core_poly = core_poly if core_poly else get_cpu_core_speed_poly()
    max_core_speed = core_poly(1.0)

    if target_core_speed > max_core_speed:
        warn_msg = (
            "Target speed ({}) is higher than "
            "the host's max core speed ({})"
        ).format(target_core_speed, max_core_speed)

        _logger.warning(warn_msg)
        warnings.warn(warn_msg, Warning)

    _logger.debug("Solving %s for %s", core_poly, target_core_speed)

    coeffs = core_poly.coeffs.copy()
    coeffs[-1] -= target_core_speed
    roots = numpy.roots(coeffs)

    _logger.debug("Poly roots: %s", roots)

    cpus = float(numpy.real(roots[numpy.isreal(roots)][0]))
    cpus = max(cpus, 1e-2)

    _logger.debug(
        "Target CPU core speed: %s ~ CPU core scale factor: %s",
        target_core_speed, cpus)

    return cpus



import logging
import pprint
import re
import signal
import time

import docker
import netifaces
import sh

from wotemu.utils import (get_current_container_id, ping_docker,
                          strip_ansi_codes)

_IFACE_LO = "lo"
_LEVEL_DEBUG = "DEBU["
_LEVEL_INFO = "INFO["

_logger = logging.getLogger(__name__)


def _find_chaos_interface():
    # Chaos commands are meant to be run in a "gateway" container.
    # These are attached to a single network, and thus there is only one interface.

    gws = netifaces.gateways()
    gw_default = gws.get("default", None)
    gw_iface = gw_default[netifaces.AF_INET][1] if gw_default else None

    if gw_iface:
        _logger.debug("Default gateway interface: %s", gw_iface)
    else:
        _logger.warning("Could not find gateway interface")

    candidates = [
        iface for iface in netifaces.interfaces()
        if iface not in [_IFACE_LO, gw_iface]
    ]

    if len(candidates) > 1:
        _logger.warning("More than 1 candidate interface: %s", candidates)

    if len(candidates) == 0:
        raise Exception("Could not find chaos network interface")

    return candidates[0]


def _pumba_log_level(line):
    if line.startswith(_LEVEL_DEBUG):
        return logging.DEBUG
    elif line.startswith(_LEVEL_INFO):
        return logging.INFO
    else:
        return logging.WARNING


def _build_out(cmd_id):
    def out(line):
        line = strip_ansi_codes(line.strip())
        _logger.log(_pumba_log_level(line), "[%s]\n\t%s", cmd_id, line.strip())

    return out


def _done(cmd, success, exit_code):
    _logger.log(
        logging.INFO if success else logging.WARNING,
        "[%s] Exit code: %s",
        cmd.ran,
        exit_code)


def create_chaos(conf, docker_url, netem, duration):
    ping_docker(docker_url=docker_url)

    docker_client = docker.DockerClient(base_url=docker_url)
    cid = get_current_container_id()
    container = docker_client.containers.get(cid)

    cmd_base = "--host {} --log-level debug netem --duration {} --interface {}".format(
        docker_url,
        duration,
        _find_chaos_interface())

    cmds = [
        "{} {} {}".format(cmd_base, netem_cmd, container.name)
        for netem_cmd in netem
    ]

    _logger.info("Running Pumba commands:\n%s", pprint.pformat(cmds))

    sh_pumba = sh.Command("pumba")

    procs = [
        sh_pumba(
            cmd.split(),
            _err_to_out=True,
            _out=_build_out(cmd_id=netem[idx]),
            _bg=True,
            _done=_done)
        for idx, cmd in enumerate(cmds)
    ]

    def exit_handler(signum, frame):
        _logger.info("Terminating Pumba processes")
        [proc.terminate() for proc in procs]

    signal.signal(signal.SIGINT, exit_handler)
    signal.signal(signal.SIGTERM, exit_handler)

    _logger.debug("Waiting for Pumba processes")
    [proc.wait() for proc in procs]
